require "forwardable"
require "logger"
require "optparse"
require "ostruct"
require "pathname"

require "bundler/setup"
require "curses"

require_relative "core_ext/hash"
require_relative "core_ext/pathname"

require_relative "text_editor/component"

require_relative "text_editor/buffer"
require_relative "text_editor/cursor"
require_relative "text_editor/keyboard"
require_relative "text_editor/logger"
require_relative "text_editor/reactor"
require_relative "text_editor/state"
require_relative "text_editor/terminal"
require_relative "text_editor/window"

require_relative "text_editor/clipboard"
require_relative "text_editor/clipboard/cygwin"
require_relative "text_editor/clipboard/linux"
require_relative "text_editor/clipboard/mac"
require_relative "text_editor/clipboard/unknown"
require_relative "text_editor/clipboard/windows"

require_relative "text_editor/command"
require_relative "text_editor/command/append_line_and_insert"
require_relative "text_editor/command/backspace"
require_relative "text_editor/command/change_mode"
require_relative "text_editor/command/cursor_down"
require_relative "text_editor/command/cursor_end"
require_relative "text_editor/command/cursor_end_and_insert"
require_relative "text_editor/command/cursor_left"
require_relative "text_editor/command/cursor_left_with_wrap"
require_relative "text_editor/command/cursor_right"
require_relative "text_editor/command/cursor_right_and_insert"
require_relative "text_editor/command/cursor_right_with_wrap"
require_relative "text_editor/command/cursor_start"
require_relative "text_editor/command/cursor_start_without_whitespace"
require_relative "text_editor/command/cursor_up"
require_relative "text_editor/command/forward_word"
require_relative "text_editor/command/insert_char"
require_relative "text_editor/command/insert_line"
require_relative "text_editor/command/insert_tab"
require_relative "text_editor/command/prepend_line_and_insert"
require_relative "text_editor/command/quit"
require_relative "text_editor/command/replace_line"
require_relative "text_editor/command/resolver"
require_relative "text_editor/command/write_file"

require_relative "text_editor/configuration"
require_relative "text_editor/configuration/bootstrap"
require_relative "text_editor/configuration/dsl"
require_relative "text_editor/configuration/flavor"
require_relative "text_editor/configuration/log"
require_relative "text_editor/configuration/mode"
require_relative "text_editor/configuration/operating_system"
require_relative "text_editor/configuration/resolver"
require_relative "text_editor/configuration/tab"

# Dependency injection container for the misc
# components required to run the text editor.
class TextEditor
  extend Configuration::DSL

  GEMSPEC = File.expand_path("../../text_editor.gemspec", __FILE__)

  attr_writer :mode

  def initialize(args = [])
    @args = args
  end

  def bootstrap
    Configuration::Bootstrap.new(self).configure(@args)
  end

  def clipboard
    @clipboard ||= Clipboard.new(self)
  end

  def command
    @command ||= Command::Resolver.new(self)
  end

  def config
    @config ||= Configuration.new(self)
  end

  def flavor
    @flavor ||= flavors[config.default_flavor]
  end

  def flavors
    @flavors ||= {}
  end

  def gemspec
    @gemspec ||= Gem::Specification.load(GEMSPEC)
  end

  def keyboard
    @keyboard ||= Keyboard.new(self)
  end

  def logger
    @logger ||= Logger.new(self)
  end

  def mode
    @mode ||= flavor.modes.values.first # FIXME
  end

  def os
    @os ||= Configuration::OperatingSystem.new(ruby.host_os)
  end

  def reactor
    @reactor ||= Reactor.new(self)
  end

  def root
    @root ||= Pathname.new(__FILE__).dirname.dirname
  end

  def ruby
    @ruby ||= OpenStruct.new(RbConfig::CONFIG)
  end

  def start
    bootstrap
    terminal.render { reactor.start }
  end

  def terminal
    @terminal ||= Terminal.new(self)
  end

  def window
    @window ||= Window.new(self)
  end

  def version
    gemspec.version.to_s
  end
end
