class TextEditor
  # An internal representation of a buffer.
  # These buffers are rendered out to the
  # terminal by {Window} instances.
  class Buffer
    using CoreExt::Pathname

    def initialize(lines = nil)
      @lines = parse_lines(lines)
    end

    def delete_char(line, column)
      validate_line!(line)
      content = self.line(line)
      content.slice!(column)
      replace_line(line, content)
    end

    def delete_line(line)
      lines = @lines.dup
      lines.delete_at(line)
      self.class.new(lines)
    end

    def each_line(&block)
      @lines.each(&block)
    end

    def insert_char(line, column, char)
      validate_line!(line)
      content = self.line(line)
      content.insert(column, char)
      replace_line(line, content)
    end

    def insert_line(line, content = "")
      lines = @lines.dup.insert(line, content)
      self.class.new(lines.compact)
    end

    def line(line)
      @lines[line].to_s.dup
    end

    def replace_char(line, column, char)
      validate_line!(line)
      content = self.line(line)
      content[column] = char
      replace_line(line, content)
    end

    def replace_line(line, content = "")
      validate_line!(line)
      lines = @lines.dup
      lines[line] = content
      self.class.new(lines.compact)
    end

    def size
      @lines.size
    end

    def to_s
      @lines.join("\n")
    end

    private

    def parse_lines(lines)
      case lines
      when nil, ""
        [""]
      when Array
        lines
      when Pathname
        parse_lines(lines.safe_read)
      when String
        lines.each_line.map(&:chomp)
      else
        raise ArgumentError
      end
    end

    def validate_line!(line)
      raise IndexError if line >= size
    end
  end
end
